import { NextApiResponse, NextApiRequest } from 'next';
import { ApiLogger, DefaultApiLoggerOption } from './api-logger.js';
import { DEFAULT_MIDDLEWARE_ROUTER_METHOD } from './constants.js';

/**
 *  an utility type to shorten writing Record<string, T = unknown>
 */
type TypedObject<T = unknown> = Record<string, T>;
/**
 *  a standard next.js api handler but with req.middleware available
 *  see [official doc](https://nextjs.org/docs/api-routes/introduction) for more details
 */
type NextApiHandlerWithMiddleware<T = unknown, M extends TypedObject = TypedObject> = (req: NextApiRequestWithMiddleware<M>, res: NextApiResponse) => T | Promise<T> | void;
/**
 *  a standard next.js api request but with req.middleware available
 */
interface NextApiRequestWithMiddleware<M extends TypedObject = TypedObject> extends NextApiRequest {
    middleware: M;
}
/**
 *  all api response combines success response & error response
 */
type ApiResponse<T = unknown> = SuccessApiResponse<T> | ErrorApiResponse;
/**
 *  api response that has http status code 2xx with data
 */
type SuccessApiResponse<T> = {
    success: true;
    data: T;
};
/**
 *  api response that has http status code 4xx or 5xx with error message
 */
type ErrorApiResponse = {
    success: false;
    message: string;
};
/**
 * RouterBuilder options
 */
type RouterBuilderOptions = Partial<{
    error: ApiErrorHandler;
    showMessage: boolean;
    logger: ApiLogger;
    loggerOption: DefaultApiLoggerOption;
}>;
type InternalMiddlewareMap = Partial<Record<MiddlewareRouterMethod, NextApiHandlerWithMiddleware<TypedObject | void, any>[]>>;
type RouterMethod = 'GET' | 'HEAD' | 'PATCH' | 'OPTIONS' | 'CONNECT' | 'DELETE' | 'TRACE' | 'POST' | 'PUT';
type MiddlewareRouterMethod = typeof DEFAULT_MIDDLEWARE_ROUTER_METHOD | RouterMethod;
type AddMiddleware<R> = {
    <T extends TypedObject | void = TypedObject, M extends TypedObject = TypedObject>(method: MiddlewareRouterMethod, handler: NextApiHandlerWithMiddleware<T, M>): R;
    <T extends TypedObject | void = TypedObject, M extends TypedObject = TypedObject>(handler: NextApiHandlerWithMiddleware<T, M>): R;
};

type ApiErrorHandler = (req: NextApiRequest, res: NextApiResponse<ErrorApiResponse>, error: Error) => void;
declare const makeErrorHandler: (showMessage: boolean) => ApiErrorHandler;

export { ApiResponse as A, ErrorApiResponse as E, InternalMiddlewareMap as I, MiddlewareRouterMethod as M, NextApiHandlerWithMiddleware as N, RouterBuilderOptions as R, SuccessApiResponse as S, TypedObject as T, NextApiRequestWithMiddleware as a, RouterMethod as b, AddMiddleware as c, ApiErrorHandler as d, makeErrorHandler as m };
